#           GOMPERTZ STATE SPACE WITH REPLICATED SAMPLING
#
#  GSS-RS model:  program to calculate ML or REML estimates & profile
#  log-likelihoods.  The program assumes an equal number (p) of
#  sampling replicates per sampling occasion.
#
#  The model is:
#       X(t) = a + c*X(t-1) + E(t),
#       Y(t) = j*X(t) + F(t).
#  Here:
#       X(t) is log-abundance of a population at time t
#         (t=0,1,2,...,q).
#       E(t) ~ normal with mean 0, variance ssq, and no 
#          autocorrelation.
#       Y(t) is a pX1 vector of estimates of X(t) obtained by
#          replicated sampling.
#       j is a pX1 vector of ones.
#       F(t) ~ multivariate normal with mean vector 0 (pX1) and var-cov
#          matrix tsq*I, where I is a pXp identity matrix.
#       a>0, ssq>0, tsq>0, -1<c<+1 are model parameters.
#
#  Be patient.  R is slow.

#----------------------------------------------------------------------
#        USER INPUT SECTION
#----------------------------------------------------------------------
#  User must supply initial parameter values here.
a0=.4;             #  Value of "a" to be used.
c0=.8;             #  Value of "c" to be used.
ssq0=.1;            #  Value of "ssq" to be used.
tsq0=.2;            #  Value of "tsq" to be used.

#  User specifies number of sampling replicates here.
p=2;     #  Number of replicates

#  User provides data here.  Statements can be altered to input data
#  from a file.  Result must be a pX(q+1) matrix YP.t, containing
#  replicated observations (log-scale) as columns.
O.t=c(24,24,9,14,10,18,7,9,9,13,13,10,5,3,13,4,10,4,5,7,3,10,5,9,
  17,8,22,19,5,10,5,5,3,4,1,2,3,5,7,6,7,11,8,8,5,7,4,3,9,15,7,9,
  14,24,3,6,6,7,11,10,8,9);  #  (24,24) are the observations at
                             #    time 0, (9,14) are the observations
                             #    at time 1, etc.
qplus1=length(O.t)/p;
q=qplus1-1;
YP.t=matrix(log(O.t),p,qplus1,byrow=FALSE);  #  Columns are replicated
                                             #    samples.

#  User sets parameter intervals for profile plots and total
#  number of horizontal axis increments here.
aalo=0.4;     #  Low value of "a"
aahi=1.9;     #  High value of "a", etc.
cclo=0.1;
cchi=0.8;
ssqlo=0.1;
ssqhi=0.4;
tsqlo=0.1;
tsqhi=0.21;
nincs=100;   #  Number of increments for the profile plots.

#----------------------------------------------------------------------
#        PROGRAM INITIALIZATION SECTION
#----------------------------------------------------------------------  
library(MASS);  #  loads miscellaneous functions (ginv, etc.).

#  Sets parameter values for profile likelihoods.
aavals=seq(aalo,aahi,by=((aahi-aalo)/nincs));
ccvals=seq(cclo,cchi,by=((cchi-cclo)/nincs));
ssqvals=seq(ssqlo,ssqhi,by=((ssqhi-ssqlo)/nincs));
tsqvals=seq(tsqlo,tsqhi,by=((tsqhi-tsqlo)/nincs));

#  These vectors will eventually hold the profiles.
profileRS.aa=aavals;
profileRS.cc=ccvals;
profileRS.ssq=ssqvals;
profileRS.tsq=tsqvals;

#  Matrices for calculating multivariate normal likelihood for REML
#  estimates.
YP.reml=matrix(YP.t,p*qplus1,1);  #  Replicated samples "stacked" in a vector.
J.p=matrix(1,p,p);          #  pXp matrix of ones.
I.p=diag(qplus1);           #  (q+1)X(q+1) identity matrix.
D.reml=kronecker(I.p,J.p);  #  kronecker product (block diagonal).
I.temp=cbind(matrix(0,q*p,p),diag(q*p));
I.temp=rbind(I.temp,matrix(0,p,p*qplus1));
D.reml=-D.reml+2*I.temp;
D.reml=D.reml[1:(q*p),];    #  This is the D transformation matrix
                            #   for REML.
WP.t=D.reml%*%YP.reml;      #  The REML-transformed observations.
j.p=matrix(1,p,1);          #  pX1 column vector of ones.
j.pXqp1=matrix(1,p*qplus1,1);  #  p*(q+1) X 1 col;umn vector of ones.
C.ml=kronecker(I.p,j.p);    #  C matrix in the multivariate normal
                            #    distribution of YP.reml.

#  Log-likelihood for ML estimation.
negloglikeRS.ml=function(theta,yt)  
{
   p=nrow(yt);
   aa=exp(theta[1]);             #  Constrains a > 0.             
   cc=2*exp(-exp(theta[2]))-1;   #  Constrains -1 < c < 1.
   sigmasq=exp(theta[3]);        #  Constrains ssq > 0. 
   tausq=exp(theta[4]);          #  Constrains tsq > 0.
   q=ncol(yt)-1;
   jj=matrix(1,p,1);
   JJ=matrix(1,p,p);
   Itausq=matrix(0,p,p);
   diag(Itausq)=rep(tausq);
   mu.t=aa/(1-cc);
   psq.t=sigmasq/(1-cc*cc);
   mm.t=jj*mu.t;
   VV.t=JJ*psq.t+Itausq;
   VVinv=ginv(VV.t);
   lnf=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
      0.5*(yt[,1]-t(mm.t))%*%VVinv%*%t(yt[,1]-t(mm.t));
   ofn=lnf;
   for (tt in 1:q)
   {
      mu.t=aa+cc*(mu.t+psq.t*t(jj)%*%VVinv%*%t(yt[,tt]-t(mm.t)));
      psq.t=cc*cc*psq.t*(1-psq.t*t(jj)%*%VVinv%*%jj)+sigmasq;
      mm.t=jj*mu.t[1];
      VV.t=JJ*psq.t[1]+Itausq;
      VVinv=ginv(VV.t);
      lnftemp=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
         0.5*(yt[,tt+1]-t(mm.t))%*%ginv(VV.t)%*%t(yt[,tt+1]-t(mm.t));
      ofn=ofn+lnftemp;
   }
   return(ofn);
}

#  Log-likelihoods for profiles:  first the parameter "a" is
#    fixed, then "c" is fixed, and so on.

#  "a" is a vector of fixed values.
negloglikeRS.a.ml=function(theta,parval,yt)  
{
   p=nrow(yt);
   aa=parval;             
   cc=2*exp(-exp(theta[1]))-1;   #  Constrains -1 < c < 1.
   sigmasq=exp(theta[2]);      #  Constrains ssq > 0. 
   tausq=exp(theta[3]);        #  Constrains tsq > 0.
   q=ncol(yt)-1;
   jj=matrix(1,p,1);
   JJ=matrix(1,p,p);
   Itausq=matrix(0,p,p);
   diag(Itausq)=rep(tausq);
   mu.t=aa/(1-cc);
   psq.t=sigmasq/(1-cc*cc);
   mm.t=jj*mu.t;
   VV.t=JJ*psq.t+Itausq;
   VVinv=ginv(VV.t);
   lnf=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
      0.5*(yt[,1]-t(mm.t))%*%VVinv%*%t(yt[,1]-t(mm.t));
   ofn=lnf;
   for (tt in 1:q)
   {
      mu.t=aa+cc*(mu.t+psq.t*t(jj)%*%VVinv%*%t(yt[,tt]-t(mm.t)));
      psq.t=cc*cc*psq.t*(1-psq.t*t(jj)%*%VVinv%*%jj)+sigmasq;
      mm.t=jj*mu.t[1];
      VV.t=JJ*psq.t[1]+Itausq;
      VVinv=ginv(VV.t);
      lnftemp=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
         0.5*(yt[,tt+1]-t(mm.t))%*%ginv(VV.t)%*%t(yt[,tt+1]-t(mm.t));
      ofn=ofn+lnftemp;
   }
   return(ofn);
}

#  "c" is a vector of fixed values
negloglikeRS.c.ml=function(theta,parval,yt)  
{
   p=nrow(yt);
   aa=exp(theta[1]);           #  Constrains a > 0.
   cc=parval;
   sigmasq=exp(theta[2]);      #  Constrains ssq > 0. 
   tausq=exp(theta[3]);        #  Constrains tsq > 0.
   q=ncol(yt)-1;
   jj=matrix(1,p,1);
   JJ=matrix(1,p,p);
   Itausq=matrix(0,p,p);
   diag(Itausq)=rep(tausq);
   mu.t=aa/(1-cc);
   psq.t=sigmasq/(1-cc*cc);
   mm.t=jj*mu.t;
   VV.t=JJ*psq.t+Itausq;
   VVinv=ginv(VV.t);
   lnf=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
      0.5*(yt[,1]-t(mm.t))%*%VVinv%*%t(yt[,1]-t(mm.t));
   ofn=lnf;
   for (tt in 1:q)
   {
      mu.t=aa+cc*(mu.t+psq.t*t(jj)%*%VVinv%*%t(yt[,tt]-t(mm.t)));
      psq.t=cc*cc*psq.t*(1-psq.t*t(jj)%*%VVinv%*%jj)+sigmasq;
      mm.t=jj*mu.t[1];
      VV.t=JJ*psq.t[1]+Itausq;
      VVinv=ginv(VV.t);
      lnftemp=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
         0.5*(yt[,tt+1]-t(mm.t))%*%ginv(VV.t)%*%t(yt[,tt+1]-t(mm.t));
      ofn=ofn+lnftemp;
   }
   return(ofn);
}

#  "sigma-squared" is a vector of fixed values.
negloglikeRS.ssq.ml=function(theta,parval,yt)  
{
   p=nrow(yt);
   aa=exp(theta[1]);             #  Constrains a > 0.            
   cc=2*exp(-exp(theta[2]))-1;   #  Constrains -1 < c < 1.
   sigmasq=parval; 
   tausq=exp(theta[3]);          #  Constrains tsq > 0.
   q=ncol(yt)-1;
   jj=matrix(1,p,1);
   JJ=matrix(1,p,p);
   Itausq=matrix(0,p,p);
   diag(Itausq)=rep(tausq);
   mu.t=aa/(1-cc);
   psq.t=sigmasq/(1-cc*cc);
   mm.t=jj*mu.t;
   VV.t=JJ*psq.t+Itausq;
   VVinv=ginv(VV.t);
   lnf=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
      0.5*(yt[,1]-t(mm.t))%*%VVinv%*%t(yt[,1]-t(mm.t));
   ofn=lnf;
   for (tt in 1:q)
   {
      mu.t=aa+cc*(mu.t+psq.t*t(jj)%*%VVinv%*%t(yt[,tt]-t(mm.t)));
      psq.t=cc*cc*psq.t*(1-psq.t*t(jj)%*%VVinv%*%jj)+sigmasq;
      mm.t=jj*mu.t[1];
      VV.t=JJ*psq.t[1]+Itausq;
      VVinv=ginv(VV.t);
      lnftemp=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
         0.5*(yt[,tt+1]-t(mm.t))%*%ginv(VV.t)%*%t(yt[,tt+1]-t(mm.t));
      ofn=ofn+lnftemp;
   }
   return(ofn);
}

#  "tau-squared" is a vector of fixed values
negloglikeRS.tsq.ml=function(theta,parval,yt)  
{
   p=nrow(yt);
   aa=exp(theta[1]);             #  Constrains a > 0.             
   cc=2*exp(-exp(theta[2]))-1;   #  Constrains -1 < c < 1.
   sigmasq=exp(theta[3]);        #  Constrains ssq > 0. 
   tausq=parval;
   q=ncol(yt)-1;
   jj=matrix(1,p,1);
   JJ=matrix(1,p,p);
   Itausq=matrix(0,p,p);
   diag(Itausq)=rep(tausq);
   mu.t=aa/(1-cc);
   psq.t=sigmasq/(1-cc*cc);
   mm.t=jj*mu.t;
   VV.t=JJ*psq.t+Itausq;
   VVinv=ginv(VV.t);
   lnf=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
      0.5*(yt[,1]-t(mm.t))%*%VVinv%*%t(yt[,1]-t(mm.t));
   ofn=lnf;
   for (tt in 1:q)
   {
      mu.t=aa+cc*(mu.t+psq.t*t(jj)%*%VVinv%*%t(yt[,tt]-t(mm.t)));
      psq.t=cc*cc*psq.t*(1-psq.t*t(jj)%*%VVinv%*%jj)+sigmasq;
      mm.t=jj*mu.t[1];
      VV.t=JJ*psq.t[1]+Itausq;
      VVinv=ginv(VV.t);
      lnftemp=(p/2)*log(2*pi)+0.5*log(det(VV.t))+
         0.5*(yt[,tt+1]-t(mm.t))%*%ginv(VV.t)%*%t(yt[,tt+1]-t(mm.t));
      ofn=ofn+lnftemp;
   }
   return(ofn);
}

#  Log-likelihood for REML estimation.
negloglikeRS.reml=function(theta,wt,pt)  
{
   cc=2*exp(-exp(theta[1]))-1;   #  Constrains -1 < c < 1.
   sigmasq=exp(theta[2]);        #  Constrains ssq > 0. 
   tausq=exp(theta[3]);          #  Constrains tsq > 0.
   q=length(wt)/pt;
   qp1=q+1;
   Sigma.mat=(sigmasq/(1-cc*cc))*toeplitz(abs(cc)^seq(0,q,1));
   J=matrix(1,pt,pt);          #  pXp matrix of ones.
   I=diag(qp1);           #  (q+1)X(q+1) identity matrix.
   D=kronecker(I,J);     #  kronecker product (block diagonal).
   I.2=cbind(matrix(0,q*pt,pt),diag(q*pt));
   I.2=rbind(I.2,matrix(0,pt,pt*qp1));
   D=-D+2*I.2;    
   D=D[1:(q*pt),];  #  This is the D transformation matrix for REML.
   j=matrix(1,pt,1);          #  pX1 column vector of ones.
   j.2=matrix(1,pt*qp1,1);  #  p*(q+1) X 1 column vector of ones.
   C=kronecker(I,j);    #  C matrix in the multivariate normal
                        #    distribution of YP.reml.
   V=C%*%Sigma.mat%*%t(C)+tausq*diag(qp1*pt);  #  Var-cov matrix of
                                               #    YP.reml.
   Phi.mat=D%*%V%*%t(D);    #  Var-cov matrix of WP.t.
   Phiinv.mat=ginv(Phi.mat);
   llikew=-(p*q/2)*log(2*pi)-(1/2)*log(det(Phi.mat))-
      (1/2)*t(wt)%*%Phiinv.mat%*%wt;  #  REML log-likelihood.
   ofn=-llikew;
   return(ofn);
}

#----------------------------------------------------------------------
#        SECTION FOR CALCULATING PROFILE LIKELIHOODS
#----------------------------------------------------------------------

for (ii in 1:(nincs+1))
{
#  Calculate profile for "a".
GSSRSaa=optim(par=c(log(-log((c0+1)/2)),log(ssq0),log(tsq0)),
   negloglikeRS.a.ml,NULL,method="Nelder-Mead",parval=aavals[ii],yt=YP.t);
profileRS.aa[ii]=-GSSRSaa$value;

#  Calculate profile for "c".
GSSRScc=optim(par=c(log(a0),log(ssq0),log(tsq0)),
   negloglikeRS.c.ml,NULL,method="Nelder-Mead",parval=ccvals[ii],yt=YP.t);
profileRS.cc[ii]=-GSSRScc$value;

#  Calculate profile for "ssq".
GSSRSssq=optim(par=c(log(a0),log(-log((c0+1)/2)),log(tsq0)),
   negloglikeRS.ssq.ml,NULL,method="Nelder-Mead",parval=ssqvals[ii],yt=YP.t);
profileRS.ssq[ii]=-GSSRSssq$value;

#  Calculate profile for "tsq".
GSSRStsq=optim(par=c(log(a0),log(-log((c0+1)/2)),log(ssq0)),
   negloglikeRS.tsq.ml,NULL,method="Nelder-Mead",parval=tsqvals[ii],yt=YP.t);
profileRS.tsq[ii]=-GSSRStsq$value;
}

#  Sets highest profile value at zero.
profileRS.aa=profileRS.aa-max(profileRS.aa);
profileRS.cc=profileRS.cc-max(profileRS.cc);
profileRS.ssq=profileRS.ssq-max(profileRS.ssq);
profileRS.tsq=profileRS.tsq-max(profileRS.tsq);

#  Profiles plotted here.
par(cex.lab=1.5, cex.axis=1.5, lwd=2);
layout(matrix(1:4, 2, 2));
plot(aavals, profileRS.aa, type="l", lty=1, ylab="profile
   log-likelihood", xlab=expression(a));
plot(ssqvals, profileRS.ssq, type="l", lty=1, ylab="profile
   log-likelihood", xlab=expression(sigma^2));
plot(ccvals, profileRS.cc, type="l", lty=1, ylab="profile
   log-likelihood", xlab=expression(c));
plot(tsqvals, profileRS.tsq, type="l", lty=1, ylab="profile
   log-likelihood",  xlab=expression(tau^2));


#----------------------------------------------------------------------
#        SECTION FOR CALCULATING ML & REML ESTIMATES
#----------------------------------------------------------------------

GSSRSml=optim(par=c(log(a0),log(-log((c0+1)/2)),log(ssq0),log(tsq0)),
   negloglikeRS.ml,NULL,method="Nelder-Mead",yt=YP.t);
GSSRSml=c(exp(GSSRSml$par[1]),2*exp(-exp(GSSRSml$par[2]))-
   1,exp(GSSRSml$par[3]),exp(GSSRSml$par[4]),-GSSRSml$value);
a.ml=GSSRSml[1];            # These are the ML estimates.
c.ml=GSSRSml[2];            #          --
ssq.ml=GSSRSml[3];          #          --
tsq.ml=GSSRSml[4];          #          --
lnlike.ml=GSSRSml[5];       #  This is the maximized log-likelihood
                            #    value.

GSSRSreml=optim(par=c(log(-log((c0+1)/2)),log(ssq0),log(tsq0)),
   negloglikeRS.reml,NULL,method="Nelder-Mead",wt=WP.t,pt=p);
GSSRSreml=c(2*exp(-exp(GSSRSreml$par[1]))-1,exp(GSSRSreml$par[2]),
   exp(GSSRSreml$par[3]),-GSSRSreml$value);

c.reml=GSSRSreml[1];            #  These are the REML estimates.
ssq.reml=GSSRSreml[2];          #          --
tsq.reml=GSSRSreml[3];          #          --
lnlike.reml=GSSRSreml[4];       #  This is the maximized log-likelihood
                                #    value.

#  Calculate REML estimate of the parameter "a".
Sigma.mat=(ssq.reml/(1-c.reml*c.reml))*toeplitz(c.reml^seq(0,q,1));
V.mat=C.ml%*%Sigma.mat%*%t(C.ml)+tsq.reml*diag(qplus1*p);
Vinv.mat=ginv(V.mat);
a.reml=(1-c.reml)*t(j.pXqp1)%*%Vinv.mat%*%YP.reml/
   (t(j.pXqp1)%*%Vinv.mat%*%j.pXqp1);            

#  Gather up stuff for printing.
estimates.ml=c(a.ml,c.ml,ssq.ml,tsq.ml,lnlike.ml);
estimates.reml=c(a.reml,c.reml,ssq.reml,tsq.reml,lnlike.reml);
names.ml=c("a.ml","c.ml","ssq.ml","tsq.ml","lnlike.ml");
names.reml=c("a.reml","c.reml","ssq.reml","tsq.reml","lnlike.reml");
values.ml=data.frame(names.ml,estimates.ml);
values.reml=data.frame(names.reml,estimates.reml);

#  Print the ML results.
values.ml;

#  Print the REML results.
values.reml;